<?php
/**
 * FRITZ!Box API - A simple wrapper for automated changes in FRITZ!Box Web-UI
 *
 * Based on FRITZ!Box API 0.4 by: Gregor Nathanael Meyer <Gregor [at] der-meyer.de>
 * License: http://creativecommons.org/licenses/by-sa/3.0/de/ Creative Commons cc-by-sa

 */

/* A simple usage example
 *
 * require_once('fritzbox_api.class.php');
 * try {
 *   $fritz = new fritzbox_api(array('hostname' => '192.168.1.1'));
 *   $fritz->set_credentials($password, $username);
 *   $docsis_info = $fritz->doGetPage('/internet/docsis_info.lua', array('update' => 'uiInfo'));
 *   print_r($docsis_info);
 *   $fritz = null;                     // destroy the object to log out
 * }
 * catch (Exception $e)
 * {
 *   echo $e->getMessage() . "\n";      // show error message if anything failed
 * }
 *
 */

class fritzbox_api {

  const INVALID_SID = '0000000000000000';

  public $hostname;
  protected $ssl = false;       // true, if ssl is to be used
  protected $stay = false;      // stay logged in, don't logout at object destruction
  protected $blocked_until = 0; // time until login is blocked
  protected $username;          // username for login
  protected $password;          // password for login
  protected $sid = false;       // session id, created by logging in

  public function __construct($params = array()) {
    $this->stay     = isset($params['stay'])     ? $params['stay']     : false;
    $this->ssl      = isset($params['ssl'])      ? $params['ssl']      : false;
    $this->hostname = isset($params['hostname']) ? $params['hostname'] : 'fritz.box';
  }


  /* log out before destroying the session */
  public function __destruct() {
    $this->stay or $this->logout();
  }


  /* on serialization, save these attributes */
  public function __sleep() {
    return array('hostname', 'ssl', 'stay', 'blocked_until', 'sid');
  }


  /* set credentials used for login */
  public function set_credentials($password, $username = null) {
    $this->username = $username;
    $this->password = $password;
  }


  /* return base url */
  public function url($uri = '') {
    return ($this->ssl ? 'https' : 'http') . '://' . $this->hostname . $uri;
  }


  /* return session ID */
  public function getSID() {
    return $this->sid;
  }


  /* return true if we are logged in */
  public function is_logged_in() {
    return $this->sid !== false;
  }


  /* perform http request (GET/POST according to given context options) */
  public function DoHttpRequest($uri, $context_options) {

    $url = $this->url($uri);
    $context = stream_context_create($context_options);

    set_error_handler(function($errno, $errstr) { return false; } );
    $result = @file_get_contents($url, false, $context);
    restore_error_handler();

    if ($result === false) {
      $err = error_get_last();
      throw new Exception("Unable to retrieve {$context_options['http']['method']} content: {$err['message']}");
    }

    return $result;
  }


  /* do a POST request on the box */
  public function doPostForm($page, $params = array(), $timeout = 10) {

    // add the sid to the request, if it is already set
    if ($this->sid !== false)
      $params['sid'] = $this->sid;

    $context_options = array('http' =>
      array(
        'method'  => 'POST',
        'timeout' => $timeout,                  // timeout in seconds
        'header'  => 'Content-type: application/x-www-form-urlencoded',
        'content' => http_build_query($params)  // post data
      )
    );

    return $this->DoHttpRequest($page, $context_options);
  }


  /* do a GET request on the box */
  public function doGetRequest($page, $params = array(), $timeout = 10) {

    // add the sid, if there is any
    if ($this->sid !== false)
      $params['sid'] = $this->sid;

    $context_options = array('http' =>
      array(
        'method'  => 'GET',
        'timeout' => $timeout,
      )
    );
    return $this->DoHttpRequest($page . '?' . http_build_query($params), $context_options);
  }


  /* login method, handles the secured login-process                           */
  /* see http://www.avm.de/de/Extern/Technical_Note_Session_ID.pdf for details */
  protected function initSID() {

    // if we are still blocked for login, don't proceed further
    if ($this->blocked_until > time())
      throw new Exception("Login to FRITZ!Box failed - still blocked for " . ($this->blocked_until - time()) . " seconds");

    // read current status
    try {
      $output = $this->DoGetRequest('/login_sid.lua');
    }
    catch (Exception $e) {
      throw new Exception("Login to FRITZ!Box failed - {$e->GetMessage()}");
    }

    // parse result as xml
    if (($xml = @simplexml_load_string($output)) === false)
      throw new Exception("Login to FRITZ!Box failed - malformed xml data for login page");

    // if the box has no password or we are already logged in, we have a sid
    if (!$xml->SID->getName())
      throw new Exception("Login to FRITZ!Box failed - unexpected xml data structure for login page");

    if ($xml->SID != self::INVALID_SID)
      return $this->sid = (string) $xml->SID;

    // we have to login and get a new SID
    $this->sid = false;

    // are we blocked from login?
    $this->blocked_until = time() + $xml->BlockTime;
    if ($xml->BlockTime > 0)
      throw new Exception("Login to FRITZ!Box failed - blocked for {$xml->BlockTime} seconds");

    // the challenge-response magic, pay attention to the mb_convert_encoding()
    $challenge = $xml->Challenge;
    $formfields = array(
      'response' => $challenge . '-' . md5(mb_convert_encoding($challenge . '-' . $this->password, "UCS-2LE", "UTF-8"))
    );

    if (!is_null($this->username))
      $formfields['username'] = $this->username;

    // do the login
    try {
      $output = $this->doPostForm('/login_sid.lua', $formfields);
    }
    catch (Exception $e) {
      throw new Exception("Login to FRITZ!Box failed - {$e->GetMessage()}");
    }

    // extract the SID from the response
    if (($xml = @simplexml_load_string($output)) === false)
      throw new Exception("Login to FRITZ!Box failed - malformed xml data for login result");

    if (!$xml->SID->getName())
      throw new Exception("Login to FRITZ!Box failed - unable to retrieve Session ID");

    if ($xml->SID != self::INVALID_SID)
      return $this->sid = (string) $xml->SID;

    // got no sid and no transfer error occurred: must be a password error
    throw new Exception("Login to FRITZ!Box failed: wrong username or password");
  }


  /* log in to FRITZ!Box */
  public function login($password, $username=null) {

    $this->set_credentials($password, $username);
    $this->initSID();
  }


  /* logout from FRITZ!Box */
  public function logout() {

    if (!$this->is_logged_in())
      return;

    try {

      $formfields = array(
        'logout'      => '1',
        'xhr'         => '1',
        'no_sidrenew' => ''
      );
      $this->doPostForm('/index.lua', $formfields);
    }
    catch (Exception $e) { }    // ignore exceptions here. Pretend that logout always succeeds.
    $this->sid = false;
  }


  /* mid level methods */

  /* get page from FRITZ!Box. Login if session is invalid and retry retrieving page */
  public function doGetPage($page, $params = array()) {

    try {
      return $this->doGetRequest($page, $params);
    }
    catch (Exception $e) {
      $msg = $e->getMessage();
      if (preg_match("@HTTP/1\.\d+\s+403\D+@", $msg)) {

        // 403: login required
        $this->initSID();

      } elseif (preg_match("@HTTP/1\.\d+\s+404\D+@", $msg)) {

        // 404: page not existing: no retry possible
        throw $e;

      } elseif (stristr($msg, "failed to open stream") === false) {

        // no connection failure: retry would fail as well
        throw $e;

      }
    }
    // try again
    return $this->doGetRequest($page, $params);
  }

  /* high level functions: perform special requests */
}
