<?php
#!/usr/bin/php

$no_http_headers = true;
$ss_fritzbox_debug = 1;

/* display no errors */
error_reporting(0);

require_once('fritzbox_api.class.php');

class ss_fritzbox_stream {
  public $colmap;
  public $counters;
  public $channelmap;
  public $base;
  public $data;


  public function __construct($colmap, $counters) {
    $this->colmap = $colmap; // table to map column numbers from html table to column names
    $this->counters = $counters;
    $this->channelmap = [];
    $this->base = [];
  }


  // set data row for one channel
  function channel_set_data($channel, $row, $overwrite = false) {
    $this->data[$channel] = $row;
    if ($overwrite || !isset($this->base[$channel]))
      $this->base[$channel] = $this->copy_counters($row);
  }


  // initialize base counter for one channel
  function copy_counters($row) {
    return array_intersect_key($row, array_flip($this->counters));
  }


  // initialize base counter values for all channels
  function init_counter_base() {
    $this->base = array_map([$this, 'copy_counters'], $this->data);
  }


  // the first counter values after connect may contain a starting value
  // substract this base value from each value
  function adjust_docsis_counters() {
    foreach ($this->data as $channel => &$row) {
      foreach ($this->counters as $counter) {
        if (isset($this->base[$channel]))
          $row[$counter] -= $this->base[$channel][$counter];
      }
    }
  }


  // detect reconnect
  function detect_docsis_reconnect($prev) {

    foreach ($this->data as $channel => $row) {
      foreach ($this->counters as $counter) {
        // not relevant if this counter has no value
        if (!is_numeric($row[$counter]))
          continue;

        // not relevant if an additional channel was just added
        if (!isset($prev[$channel]) || !is_numeric($prev[$channel][$counter]))
          continue;

        // reconnect happened, if one of the counters went backwards
        if ($row[$counter] < $prev[$channel][$counter] + $this->base[$channel][$counter])
          return true;
      }
    }
    return false;
  }


  // process xml docsis data for one direction
  function process_docsis_data($table, $idcol) {

    $this->data = [];
    $data_newchannel = [];

    // iterate over each channel (listed as column headers) in the html table
    $num_channels = $table->tr->th->count();
    for ($channel = 1; $channel <= $num_channels; $channel++) {

      // skip channels without valid channel number
      if (!is_numeric((string)$table->tr->th[$channel]))
        continue;

      // extract data from $table for all columns from $colmap to $row[$keyword]
      $row = array_map(function($idx_row) use ($table, $channel) {
        return filter_var($table->tr[$idx_row]->{$idx_row ? 'td' : 'th'}[$channel],
                            FILTER_SANITIZE_NUMBER_FLOAT, FILTER_FLAG_ALLOW_FRACTION | FILTER_FLAG_ALLOW_THOUSAND);
      }, $this->colmap);

      // insert channel with known freq into same data slot every time
      if (is_numeric($id = $row[$idcol])) {

        if (array_key_exists($id, $this->channelmap)) {
          $this->channel_set_data($this->channelmap[$id], $row);
        } else {
          // freq not known, add $row to output table later
          $data_newchannel[$id] = $row;
        }

      }
    }

    // sort new channels by id column
    ksort($data_newchannel);

    // add new channels to table, overwriting empty rows or append new slot
    $channel = 1;
    foreach ($data_newchannel as $id => $row) {
      while (array_key_exists($channel, $this->data))
        $channel++;

      $this->channel_set_data($channel, $row, true);

      // remove map entry that previously pointed to $channel
      if (($old_id = array_search($channel, $this->channelmap)) !== false)
        unset($this->channelmap[$old_id]);

      // add new map entry for $channel
      $this->channelmap[$id] = $channel;
    }

    // fill empty data rows for vanished channels
    foreach ($this->channelmap as $channel) {
      if (!array_key_exists($channel, $this->data)) {
        $this->channel_set_data($channel, array_fill_keys(array_keys($this->colmap), ''));
        unset($this->base[$channel]);
      }
    }

    ksort($this->data);
  }


  function import_docsis_data($table, $idcol) {

    $prev = $this->data;
    $this->process_docsis_data($table, $idcol);

    // in case of a reconnect, initialize base counter values
    if ($this->detect_docsis_reconnect($prev))
      $this->init_counter_base();

    $this->adjust_docsis_counters();
  }

}

/* ------------------------------------------------------------------------- */

class ss_fritzbox extends fritzbox_api {
  public $no_access_until;
  public $dumpdir;
  public $page;
  public $data_valid_until;
  public $data;


  public function __construct() {

    parent::__construct( [ 'stay' => true, 'ssl' => false] );
    $this->no_access_until = 0;
    $this->page = [
      'docsis'    => [ 'page' => '/internet/docsis_info.lua',   'params' => ['xhr' => '1', 'update' => 'uiInfo'], 'cache' => 300],
      'smarthome' => [ 'page' => '/net/home_auto_overview.lua', 'params' => ['xhr' => '1', 'useajax' => '1' ],    'cache' => 0],
    ];
    $this->data_valid_until = [];
    $this->data = [];

    $this->data['docsis'] = [
      'ds' => new ss_fritzbox_stream([
        'channel' =>       0, // 1..24
        'channelid' =>     1, // 30, 31, 40, 41, ..
        'freq' =>          2,
        'modulation' =>    3,
        'power' =>         4,
        'mse' =>           5,
        'latency' =>       6,
        'correctable' =>   7,
        'uncorrectable' => 8
      ], [ 'correctable', 'uncorrectable' ]),

      'us' => new ss_fritzbox_stream([
        'channel' =>       0, // 1..5
        'channelid' =>     1, // 30, 31, 40, 41, ..
        'freq' =>          2,
        'modulation' =>    3,
        'multiplex' =>     4,
        'power' =>         5
      ], [])
    ];
    $this->data_valid_until['docsis'] = 0;
  }


  public function __sleep() {
    return array_merge([ 'no_access_until', 'page', 'data_valid_until', 'data' ], parent::__sleep());
  }


  function reinit($hostname, $username, $password, $debug) {

    $this->dumpdir = null;

    if ($debug & 1) {
      $this->dumpdir = sys_get_temp_dir() . '/fritzbox';
      is_dir($this->dumpdir) or mkdir($this->dumpdir, 0777, true);
    }

    if ($debug & 2) {
      $this->data_valid_until['docsis'] = 0;
      error_reporting(E_ALL);
    }

    $this->set_credentials($password, $username);
    if ($this->hostname != $hostname) {
      $this->hostname = $hostname;
      $this->sid = false;
      $this->initSID();
      $this->no_access_until = 0;
      $this->data_valid_until['docsis'] = 0;
    }

  }


  function get_page_xml($key) {

    if ($this->no_access_until >= time())
      throw new Exception('blocked');

    try {
      // get web page with requested info
      $html = $this->doGetPage($this->page[$key]['page'], $this->page[$key]['params']);

      // save html for analysis purposes
      if (isset($this->dumpdir))
        file_put_contents("{$this->dumpdir}/{$key}-" . date('Y-m-d-H-i-s') . ".html", $html);

      // extract xml
      $doc = new DOMDocument();
      $doc->loadHTML($html);
      $xml = simplexml_import_dom($doc);
      if ($xml === false)
        throw new Exception("Unable to parse FRITZ!Box {$key} info from {$this->hostname}");
    }
    catch (Exception $e) {
      // unable to retrieve page or unable to parse retrieved page: don't access
      // FRITZ!Box for the next 5 minutes (login failure) or 10 seconds (otherwise)
      $this->no_access_until = max($this->blocked_until, time() + ($this->is_logged_in() ? 10 : 300));
      throw new Exception("ERROR: {$e->getMessage()} - suspending access until " . date("d.m.Y H:i:s", $this->no_access_until));
    }
    return $xml;
  }

  function output($data, $cmd, $kw, $i) {

    switch ($cmd) {
      case 'index':
        foreach ($data as $i => $row) {
          print "{$i}\n";
        }
        return;

      case 'num_indexes':
        return count($data);

      case 'query':
        foreach ($data as $i => $row) {
          print "{$i}!{$row[$kw]}\n";
        }
        return;

      case 'get':
        return isset($data[$i]) ? $data[$i][$kw] : 'U';
    }
    throw new Exception("ERROR: invalid script query parameter: '$cmd'");
  }


  /* get docsis info from FRITZ!Box and return parsed data */
  function get_docsis_data($key, $dir) {

    // refresh cached data, if stored data is stale
    if ($this->data_valid_until[$key] < time()) {

      $xml = $this->get_page_xml($key);

      // process each direction
      // pick us and ds tables from $xml and process them
      foreach ($xml->body->div as $div) {
        foreach ($this->data[$key] as $dir1 => $stream) {
          if ($div['id'] == "uiInfo_$dir1")
            $stream->import_docsis_data($div->table, 'freq');
        }
      }

      $this->data_valid_until[$key] = floor(time()/300 + 1) * 300 - 1; // next 5 minute border
    }

    return $this->data[$key][$dir]->data;
  }


  function import_smarthome_data($table) {
    $data = [];
    foreach ($table->tr as $tr) {
      if ($tr->td->count() > 0) {
        $value = filter_var($tr->td[3], FILTER_SANITIZE_NUMBER_FLOAT, FILTER_FLAG_ALLOW_FRACTION | FILTER_FLAG_ALLOW_THOUSAND);
        $value = str_replace(',', '.', $value);
        $device = (string)$tr->td[0]['datalabel'];
        $data[str_replace([' ', '!'], '_', $device)] = [ 'device' => $device, 'temp' => $value ];
      }
    }
    return $data;
  }


  /* get smarthome info from FRITZ!Box and return parsed data */
  function get_smarthome_data($key) {

    $xml = $this->get_page_xml($key);

    foreach ($xml->body->form->div as $div) {
      if ($div['id'] == "uiSmarthomeTables") {
        return $this->import_smarthome_data($div->table);
      }
    }

  }


  function output_smarthome($cmd, $kw, $i) {
    $data = $this->get_smarthome_data('smarthome');
    return $this->output($data, $cmd, $kw, $i);
  }


  function output_docsis($dir, $cmd, $kw, $i) {
    $data = $this->get_docsis_data('docsis', $dir);
    return $this->output($data, $cmd, $kw, $i);
  }

}


/* ------------------------------------------------------------------------- */

// $debug: 1=dump html, 2=ignore cache, 3=both
function fritzbox_session_get($hostname, $host_id, $auth, $debug=1) {

  // community as password?username
  list($password, $username) = array_pad(explode('?', $auth, 2), 2, null);

  session_save_path($debug & 2 ? sys_get_temp_dir() : '/var/lib/php/session');
  ini_set('session.gc_probability', 0);
  session_id("cacti-fb-{$host_id}-" . posix_getuid());

  session_start();
  $fb =& $_SESSION['fritzbox'];
  if (!isset($fb))
    $fb = new ss_fritzbox();
  $fb->reinit($hostname, $username, $password, $debug);
  return $fb;
}


function fritzbox_session_error($msg) {
  global $called_by_script_server;

  session_write_close();

  if (isset($called_by_script_server)) {
    if ($msg != 'blocked')
      cacti_log($msg);
  } else {
    echo "{$msg}\n";
  }
}


/* get smarthome information from cable FRITZ!Box 6xxx with Fritz!OS >= 6.50 */
function ss_fritzbox_smarthome($hostname, $host_id, $auth, $cmd, $arg1='', $arg2='') {
  global $ss_fritzbox_debug;

  if (!$cmd)
    return "ERROR: usage: ss_fritzbox_smarthome(<hostname>, <host_id>, <auth>, <cmd>[, <kw>[, <idx>]])";

  try {
    $fb = fritzbox_session_get($hostname, $host_id, $auth, $ss_fritzbox_debug);
    $retval = $fb->output_smarthome($cmd, $arg1, $arg2);
    session_write_close();
    return $retval;
  }
  catch (Exception $e) {
    fritzbox_session_error($e->getMessage());
  }
  return 'U';
}


/* get docsis information from cable FRITZ!Box 6xxx with Fritz!OS >= 6.50 */
function ss_fritzbox_docsis($hostname, $host_id, $auth, $dir, $cmd, $arg1='', $arg2='') {
  global $ss_fritzbox_debug;

  if (!$cmd)
    return "ERROR: usage: ss_fritzbox_docsis(<hostname>, <host_id>, <auth>, <dir>, <cmd>[, <kw>[, <idx>]])";

  try {
    $fb = fritzbox_session_get($hostname, $host_id, $auth, $ss_fritzbox_debug);
    $retval = $fb->output_docsis($dir, $cmd, $arg1, $arg2);
    session_write_close();
    return $retval;
  }
  catch (Exception $e) {
    fritzbox_session_error($e->getMessage());
  }
  return 'U';
}


if (!isset($called_by_script_server)) {
  include_once(dirname(__FILE__) . "/../include/cli_check.php");
  array_shift($_SERVER['argv']);
  if ($_SERVER['argv'][0] == "--debug") {
    $ss_fritzbox_debug = 3;
    array_shift($_SERVER['argv']);
  }

  if (strpos($_SERVER['argv'][0], "ss_") === 0) {
    $func = array_shift($_SERVER['argv']);
  } elseif ($_SERVER['argv'][3] == "ds" || $_SERVER['argv'][3] == "us") {
    $func = "ss_fritzbox_docsis";
  } else {
    $func = "ss_fritzbox_smarthome";
  }

  print call_user_func_array($func, $_SERVER['argv']);
}
